# Stubs for xml.sax (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

class SAXException(Exception):
    def __init__(self, msg, exception=None): ...
    def getMessage(self): ...
    def getException(self): ...
    def __getitem__(self, ix): ...

class SAXParseException(SAXException):
    def __init__(self, msg, exception, locator): ...
    def getColumnNumber(self): ...
    def getLineNumber(self): ...
    def getPublicId(self): ...
    def getSystemId(self): ...

class SAXNotRecognizedException(SAXException): ...
class SAXNotSupportedException(SAXException): ...
class SAXReaderNotAvailable(SAXNotSupportedException): ...
