# Stubs for sqlalchemy.engine (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from .base import Connection as Connection
from .base import Engine as Engine
from .base import RowProxy as RowProxy
from .base import Transaction as Transaction

def create_engine(*args, **kwargs): ...
def engine_from_config(configuration, prefix=..., **kwargs): ...
