from typing import Any, List, Tuple

# Dummy until I figure out something better.
class Connectable:
    pass

class Connection:
    def begin(self): ...
    def execute(self, object, *multiparams, **params): ...

class Engine(object): ...

class RowProxy:
    def items(self) -> List[Tuple[Any, Any]]: ...
    def keys(self) -> List[Any]: ...
    def values(self) -> List[Any]: ...
    def __getitem__(self, key: str): ...

class Transaction:
    def commit(self): ...
    def rollback(self): ...
