# Stubs for pytz.lazy (Python 3.5)

from typing import Any, Iterable, List, Set, Dict  # NOQA
from collections import Mapping

class LazyDict(Mapping):
    pass

class LazyList(List):
    pass

class LazySet(Set):
    pass
